#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include "sstr.h"

void oom();

/*TODO: not string but buffer: accept length */

struct sstr {
  unsigned used;
  struct sstr *next;
  unsigned char str[1]; /* first byte is length - limited to 255 bytes */
};

static struct sstr *hash[256];
static struct sstr *last;

void sstrstat() {}

static unsigned sstrhash(const unsigned char *s) {
  /* taken from DJB's cdb_hash code */
  unsigned h = 5381;
  while(*s) {
    h += (h << 5);
    h ^= *s++;
  }
  return h & 255;
}

char *sstrdup(const char *str) {
  struct sstr *s;
  unsigned len = strlen(str);
  unsigned i;
  if (len > 255) return NULL;
  if (last && last->str[0] == len && memcmp(last->str+1, str, len) == 0) {
    ++last->used;
    return last->str + 1;
  }
  i = sstrhash(str);
  for(s = hash[i]; s; s = s->next)
    if (s->str[0] == len && memcmp(s->str+1, str, len) == 0) {
      last = s;
      ++s->used;
      return s->str + 1;
    }
  s = (struct sstr*)malloc(sizeof(*s) + len + 1);
  if (!s) {
    oom();
    return NULL;
  }
  s->str[0] = len;
  strcpy(s->str+1, str);
  s->used = 1;
  s->next = hash[i];
  hash[i] = last = s;
  return s->str+1;
}

void sstrfree(char *str) {
  struct sstr *s = (struct sstr *)(str - 1 - offsetof(struct sstr, str));
  unsigned i;
  if (!str) return;
  if (--s->used != 0) return;
  i = sstrhash(str);
  if (s == hash[i])
    hash[i] = s->next;
  else {
    struct sstr *p = hash[i];
    while(p->next != s)
      if (!p->next) abort();
      else p = p->next;
    p->next = s->next;
  }
  free(s);
  if (last == s) last = NULL;
}

char *sstrref(char *str) {
  if (str)
    ((struct sstr *)(str - 1 - offsetof(struct sstr, str)))->used++;
  return str;
}
